@@ -39,10 +39,14 @@
 
 
 #define IN_GLYPH( pos )      buffer->in_string[(pos)].gindex
-#define IN_CURGLYPH( pos )   buffer->in_string[(pos) + buffer->in_pos].gindex
+#define IN_ITEM( pos )       (&buffer->in_string[(pos)])
+#define IN_CURGLYPH()        buffer->in_string[buffer->in_pos].gindex
+#define IN_CURITEM()         (&buffer->in_string[buffer->in_pos])
 #define IN_PROPERTIES( pos ) buffer->in_string[(pos)].properties
 #define IN_LIGID( pos )      buffer->in_string[(pos)].ligID
 
+#define OUT_GLYPH( pos )     buffer->out_string[(pos)].gindex
+#define OUT_ITEM( pos )      (&buffer->out_string[(pos)])
 
 #define ADD_String( buffer, num_in, num_out, glyph_data, component, ligID )             \
           ( ( error = otl_buffer_add_output_glyphs( (buffer),                           \
@@ -326,17 +330,17 @@
     if ( context_length != 0xFFFF && context_length < 1 )
       return TTO_Err_Not_Covered;
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
-    error = Coverage_Index( &ss->Coverage, IN_CURGLYPH( 0 ), &index );
+    error = Coverage_Index( &ss->Coverage, IN_CURGLYPH(), &index );
     if ( error )
       return error;
 
     switch ( ss->SubstFormat )
     {
     case 1:
-	    value = ( IN_CURGLYPH( 0 ) + ss->ssf.ssf1.DeltaGlyphID ) & 0xFFFF;
+	    value = ( IN_CURGLYPH() + ss->ssf.ssf1.DeltaGlyphID ) & 0xFFFF;
       if ( ADD_Glyph( buffer, value, 0xFFFF, 0xFFFF ) )
         return error;
       break;
@@ -530,10 +534,10 @@
     if ( context_length != 0xFFFF && context_length < 1 )
       return TTO_Err_Not_Covered;
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
-    error = Coverage_Index( &ms->Coverage, IN_CURGLYPH( 0 ), &index );
+    error = Coverage_Index( &ms->Coverage, IN_CURGLYPH(), &index );
     if ( error )
       return error;
 
@@ -728,10 +732,10 @@
     if ( context_length != 0xFFFF && context_length < 1 )
       return TTO_Err_Not_Covered;
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
-    error = Coverage_Index( &as->Coverage, IN_CURGLYPH( 0 ), &index );
+    error = Coverage_Index( &as->Coverage, IN_CURGLYPH(), &index );
     if ( error )
       return error;
 
@@ -740,7 +744,7 @@
     /* we use a user-defined callback function to get the alternate index */
 
     if ( gsub->altfunc )
-      alt_index = (gsub->altfunc)( buffer->out_pos, IN_CURGLYPH( 0 ),
+      alt_index = (gsub->altfunc)( buffer->out_pos, IN_CURGLYPH(),
                                    aset.GlyphCount, aset.Alternate,
                                    gsub->data );
     else
@@ -1006,13 +1010,13 @@
     TTO_Ligature*  lig;
 
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
     if ( property == TTO_MARK || property & IGNORE_SPECIAL_MARKS )
       first_is_mark = TRUE;
 
-    error = Coverage_Index( &ls->Coverage, IN_CURGLYPH( 0 ), &index );
+    error = Coverage_Index( &ls->Coverage, IN_CURGLYPH(), &index );
     if ( error )
       return error;
 
@@ -1037,7 +1041,7 @@
 
       for ( i = 1, j = buffer->in_pos + 1; i < lig->ComponentCount; i++, j++ )
       {
-        while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+        while ( CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
         {
           if ( error && error != TTO_Err_Not_Covered )
             return error;
@@ -1099,9 +1103,9 @@
 
 	for ( i = 0; i < lig->ComponentCount - 1; i++ )
 	{
-	  while ( CHECK_Property( gdef, IN_CURGLYPH( 0 ),
+	  while ( CHECK_Property( gdef, IN_CURITEM(),
 				  flags, &property ) )
-	    if ( ADD_Glyph( buffer, IN_CURGLYPH( 0 ),
+	    if ( ADD_Glyph( buffer, IN_CURGLYPH(),
 			    i, ligID ) )
 	      return error;
 
@@ -1155,7 +1159,7 @@
         {
           /* XXX "can't happen" -- but don't count on it */
 
-          if ( ADD_Glyph( buffer, IN_CURGLYPH( 0 ),
+          if ( ADD_Glyph( buffer, IN_CURGLYPH(),
 			  0xFFFF, 0xFFFF ) )
             return error;
           i++;
@@ -1167,7 +1171,7 @@
       {
         /* No substitution for this index */
 
-        if ( ADD_Glyph( buffer, IN_CURGLYPH( 0 ),
+        if ( ADD_Glyph( buffer, IN_CURGLYPH(),
 			0xFFFF, 0xFFFF ) )
           return error;
         i++;
@@ -1907,10 +1911,10 @@
 
     gdef = gsub->gdef;
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
-    error = Coverage_Index( &csf1->Coverage, IN_CURGLYPH( 0 ), &index );
+    error = Coverage_Index( &csf1->Coverage, IN_CURGLYPH(), &index );
     if ( error )
       return error;
 
@@ -1927,7 +1931,7 @@
 
       for ( i = 1, j = buffer->in_pos + 1; i < sr[k].GlyphCount; i++, j++ )
       {
-        while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+        while ( CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
         {
           if ( error && error != TTO_Err_Not_Covered )
             return error;
@@ -1976,21 +1980,21 @@
 
     gdef = gsub->gdef;
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
     /* Note: The coverage table in format 2 doesn't give an index into
              anything.  It just lets us know whether or not we need to
              do any lookup at all.                                     */
 
-    error = Coverage_Index( &csf2->Coverage, IN_CURGLYPH( 0 ), &index );
+    error = Coverage_Index( &csf2->Coverage, IN_CURGLYPH(), &index );
     if ( error )
       return error;
 
     if ( ALLOC_ARRAY( classes, csf2->MaxContextLength, FT_UShort ) )
       return error;
 
-    error = Get_Class( &csf2->ClassDef, IN_CURGLYPH( 0 ),
+    error = Get_Class( &csf2->ClassDef, IN_CURGLYPH(),
                        &classes[0], NULL );
     if ( error && error != TTO_Err_Not_Covered )
       goto End;
@@ -2019,7 +2023,7 @@
 
       for ( i = 1, j = buffer->in_pos + 1; i < sr->GlyphCount; i++, j++ )
       {
-        while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+        while ( CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
         {
           if ( error && error != TTO_Err_Not_Covered )
             goto End;
@@ -2078,7 +2082,7 @@
 
     gdef = gsub->gdef;
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
     if ( context_length != 0xFFFF && context_length < csf3->GlyphCount )
@@ -2091,7 +2095,7 @@
 
     for ( i = 1, j = buffer->in_pos + 1; i < csf3->GlyphCount; i++, j++ )
     {
-      while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+      while ( CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
       {
         if ( error && error != TTO_Err_Not_Covered )
           return error;
@@ -3183,10 +3187,10 @@
 
     gdef = gsub->gdef;
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
-    error = Coverage_Index( &ccsf1->Coverage, IN_CURGLYPH( 0 ), &index );
+    error = Coverage_Index( &ccsf1->Coverage, IN_CURGLYPH(), &index );
     if ( error )
       return error;
 
@@ -3205,7 +3209,7 @@
 
       /* check whether context is too long; it is a first guess only */
 
-      if ( bgc > buffer->in_pos || buffer->in_pos + igc + lgc > buffer->in_length )
+      if ( bgc > buffer->out_pos || buffer->in_pos + igc + lgc > buffer->in_length )
 	goto next_chainsubrule;
 
       if ( bgc )
@@ -3213,9 +3217,9 @@
         /* since we don't know in advance the number of glyphs to inspect,
            we search backwards for matches in the backtrack glyph array    */
 
-        for ( i = 0, j = buffer->in_pos - 1; i < bgc; i++, j-- )
+        for ( i = 0, j = buffer->out_pos - 1; i < bgc; i++, j-- )
         {
-          while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+          while ( CHECK_Property( gdef, OUT_ITEM( j ), flags, &property ) )
           {
             if ( error && error != TTO_Err_Not_Covered )
               return error;
@@ -3235,7 +3239,7 @@
                Backtrack offsets -  3  2  1  0
                Lookahead offsets -                    0  1  2  3           */
 
-          if ( IN_GLYPH( j ) != curr_csr.Backtrack[i] )
+          if ( OUT_GLYPH( j ) != curr_csr.Backtrack[i] )
 	    goto next_chainsubrule;
         }
       }
@@ -3244,7 +3248,7 @@
 
       for ( i = 1, j = buffer->in_pos + 1; i < igc; i++, j++ )
       {
-        while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+        while ( CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
         {
           if ( error && error != TTO_Err_Not_Covered )
             return error;
@@ -3263,7 +3267,7 @@
 
       for ( i = 0; i < lgc; i++, j++ )
       {
-        while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+        while ( CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
         {
           if ( error && error != TTO_Err_Not_Covered )
             return error;
@@ -3324,14 +3328,14 @@
     gdef = gsub->gdef;
     memory = gsub->memory;
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
     /* Note: The coverage table in format 2 doesn't give an index into
              anything.  It just lets us know whether or not we need to
              do any lookup at all.                                     */
 
-    error = Coverage_Index( &ccsf2->Coverage, IN_CURGLYPH( 0 ), &index );
+    error = Coverage_Index( &ccsf2->Coverage, IN_CURGLYPH(), &index );
     if ( error )
       return error;
 
@@ -3347,7 +3351,7 @@
       goto End2;
     known_lookahead_classes = 0;
 
-    error = Get_Class( &ccsf2->InputClassDef, IN_CURGLYPH( 0 ),
+    error = Get_Class( &ccsf2->InputClassDef, IN_CURGLYPH(),
                        &input_classes[0], NULL );
     if ( error && error != TTO_Err_Not_Covered )
       goto End1;
@@ -3371,7 +3375,7 @@
 
       /* check whether context is too long; it is a first guess only */
 
-      if ( bgc > buffer->in_pos || buffer->in_pos + igc + lgc > buffer->in_length )
+      if ( bgc > buffer->out_pos || buffer->in_pos + igc + lgc > buffer->in_length )
 	goto next_chainsubclassrule;
 
       if ( bgc )
@@ -3382,9 +3386,9 @@
 
         bc       = ccsr.Backtrack;
 
-        for ( i = 0, j = buffer->in_pos - 1; i < bgc; i++, j-- )
+        for ( i = 0, j = buffer->out_pos - 1; i < bgc; i++, j-- )
         {
-          while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+          while ( CHECK_Property( gdef, OUT_ITEM( j ), flags, &property ) )
           {
             if ( error && error != TTO_Err_Not_Covered )
               goto End1;
@@ -3398,7 +3402,7 @@
           {
             /* Keeps us from having to do this for each rule */
 
-            error = Get_Class( &ccsf2->BacktrackClassDef, IN_GLYPH( j ),
+            error = Get_Class( &ccsf2->BacktrackClassDef, OUT_GLYPH( j ),
                                &backtrack_classes[i], NULL );
             if ( error && error != TTO_Err_Not_Covered )
               goto End1;
@@ -3416,7 +3420,7 @@
 
       for ( i = 1, j = buffer->in_pos + 1; i < igc; i++, j++ )
       {
-        while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+        while ( CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
         {
           if ( error && error != TTO_Err_Not_Covered )
             goto End1;
@@ -3446,7 +3450,7 @@
 
       for ( i = 0; i < lgc; i++, j++ )
       {
-        while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+        while ( CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
         {
           if ( error && error != TTO_Err_Not_Covered )
             goto End1;
@@ -3514,7 +3518,7 @@
 
     gdef = gsub->gdef;
 
-    if ( CHECK_Property( gdef, IN_CURGLYPH( 0 ), flags, &property ) )
+    if ( CHECK_Property( gdef, IN_CURITEM(), flags, &property ) )
       return error;
 
     bgc = ccsf3->BacktrackGlyphCount;
@@ -3526,7 +3530,7 @@
 
     /* check whether context is too long; it is a first guess only */
 
-    if ( bgc > buffer->in_pos || buffer->in_pos + igc + lgc > buffer->in_length )
+    if ( bgc > buffer->out_pos || buffer->in_pos + igc + lgc > buffer->in_length )
       return TTO_Err_Not_Covered;
 
     if ( bgc )
@@ -3536,9 +3540,9 @@
 
       bc       = ccsf3->BacktrackCoverage;
 
-      for ( i = 0, j = buffer->in_pos - 1; i < bgc; i++, j-- )
+      for ( i = 0, j = buffer->out_pos - 1; i < bgc; i++, j-- )
       {
-        while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+        while ( CHECK_Property( gdef, OUT_ITEM( j ), flags, &property ) )
         {
           if ( error && error != TTO_Err_Not_Covered )
             return error;
@@ -3548,7 +3552,7 @@
 	  j--;
         }
 
-        error = Coverage_Index( &bc[i], IN_GLYPH( j ), &index );
+        error = Coverage_Index( &bc[i], OUT_GLYPH( j ), &index );
         if ( error )
           return error;
       }
@@ -3559,7 +3563,7 @@
     for ( i = 0, j = buffer->in_pos; i < igc; i++, j++ )
     {
       /* We already called CHECK_Property for IN_GLYPH( buffer->in_pos ) */
-      while ( j > buffer->in_pos && CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+      while ( j > buffer->in_pos && CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
       {
         if ( error && error != TTO_Err_Not_Covered )
           return error;
@@ -3581,7 +3585,7 @@
 
     for ( i = 0; i < lgc; i++, j++ )
     {
-      while ( CHECK_Property( gdef, IN_GLYPH( j ), flags, &property ) )
+      while ( CHECK_Property( gdef, IN_ITEM( j ), flags, &property ) )
       {
         if ( error && error != TTO_Err_Not_Covered )
           return error;
@@ -4016,7 +4020,7 @@
   {
     FT_Error  error, retError = TTO_Err_Not_Covered;
 
-    FT_UShort*  properties = gsub->LookupList.Properties;
+    FT_UInt*  properties = gsub->LookupList.Properties;
 
     int      nesting_level = 0;
 
@@ -4039,8 +4043,8 @@
       else
         error = TTO_Err_Not_Covered;
 
-      if ( error == TTO_Err_Not_Covered ) 
-        if ( ADD_Glyph( buffer, IN_CURGLYPH( 0 ), 0xFFFF, 0xFFFF ) )
+      if ( error == TTO_Err_Not_Covered )
+        if ( otl_buffer_copy_output_glyph ( buffer ) )
           return error;
     }
 
@@ -4051,19 +4055,23 @@
   EXPORT_FUNC
   FT_Error  TT_GSUB_Add_Feature( TTO_GSUBHeader*  gsub,
                                  FT_UShort        feature_index,
-                                 FT_UShort        property )
+                                 FT_UInt          property )
   {
     FT_UShort    i;
 
     TTO_Feature  feature;
-    FT_UShort*   properties;
+    FT_UInt*     properties;
     FT_UShort*   index;
 
-
+    /* Each feature can only be added once once */
+    
     if ( !gsub ||
-         feature_index >= gsub->FeatureList.FeatureCount )
+         feature_index >= gsub->FeatureList.FeatureCount ||
+	 gsub->FeatureList.ApplyCount == gsub->FeatureList.FeatureCount )
       return TT_Err_Invalid_Argument;
 
+    gsub->FeatureList.ApplyOrder[gsub->FeatureList.ApplyCount++] = feature_index;
+
     properties = gsub->LookupList.Properties;
 
     feature = gsub->FeatureList.FeatureRecord[feature_index].Feature;
@@ -4081,12 +4089,14 @@
   {
     FT_UShort i;
 
-    FT_UShort*  properties;
+    FT_UInt*  properties;
 
 
     if ( !gsub )
       return TT_Err_Invalid_Argument;
 
+    gsub->FeatureList.ApplyCount = 0;
+
     properties = gsub->LookupList.Properties;
 
     for ( i = 0; i < gsub->LookupList.LookupCount; i++ )
@@ -4116,32 +4126,34 @@
 				  OTL_Buffer        buffer )
   {
     FT_Error          error, retError = TTO_Err_Not_Covered;
-    FT_UShort         j;
-
-    FT_UShort*        properties;
+    FT_UShort         i, j, feature_index;
+    TTO_Feature       feature;
 
     if ( !gsub ||
          !buffer || buffer->in_length == 0 || buffer->in_pos >= buffer->in_length )
       return TT_Err_Invalid_Argument;
 
-    properties = gsub->LookupList.Properties;
-    
-    for ( j = 0; j < gsub->LookupList.LookupCount; j++ )
-      if ( properties[j] )
+    for ( i = 0; i < gsub->FeatureList.ApplyCount; i++)
+    {
+      feature_index = gsub->FeatureList.ApplyOrder[i];
+      feature = gsub->FeatureList.FeatureRecord[feature_index].Feature;
+
+      for ( j = 0; j < feature.LookupListCount; j++ )
       {
-        error = Do_String_Lookup( gsub, j, buffer );
-        if ( error )
+        error = Do_String_Lookup( gsub, feature.LookupListIndex[j], buffer );
+	if ( error )
 	{
 	  if ( error != TTO_Err_Not_Covered )
 	    goto End;
 	}
 	else
 	  retError = error;
-	  
+        
 	error = otl_buffer_swap( buffer );
 	if ( error )
 	  goto End;
-      }
+      } 
+    }
     
     error = retError;
 
