@@ -20,6 +20,7 @@
  */
 
 #include <errno.h>
+#include <math.h>
 #include <string.h>
 #include <stdlib.h>
 
@@ -986,6 +987,166 @@ pango_parse_stretch (const char   *str,
   return FALSE;
 }
 
+GType
+pango_matrix_get_type (void)
+{
+  static GType our_type = 0;
+  
+  if (our_type == 0)
+    our_type = g_boxed_type_register_static ("PangoMatrix",
+					     (GBoxedCopyFunc) pango_matrix_copy,
+					     (GBoxedFreeFunc) pango_matrix_free);
+
+  return our_type;
+}
+
+/**
+ * pango_matrix_copy:
+ * @matrix: a #PangoMatrix
+ * 
+ * Copies a #PangoMatrix.
+ * 
+ * Return value: a copy of @matrix. The result must be freed with
+ *  pango_matrix_free().
+ *
+ * Since: 1.6
+ **/
+PangoMatrix *
+pango_matrix_copy (const PangoMatrix *matrix)
+{
+  g_return_val_if_fail (matrix != NULL, NULL);
+
+  return g_memdup (matrix, sizeof (PangoMatrix));
+}
+
+/**
+ * pango_matrix_free:
+ * @matrix: a #PangoMatrix
+ * 
+ * Free a #PangoMatrix created with pango_matrix_copy().
+ *
+ * Since: 1.6
+ **/
+void
+pango_matrix_free (PangoMatrix *matrix)
+{
+  g_return_if_fail (matrix != NULL);
+  
+  g_free (matrix);
+}
+
+/**
+ * pango_matrix_translate:
+ * @matrix: a #PangoMatrix
+ * @tx: amount to translate in the X direction
+ * @ty: amount to translate in the Y direction
+ * 
+ * Changes the transformation represented by @matrix to be the
+ * transformation given by first translating by (@tx, @ty)
+ * then applying the original transformation.
+ *
+ * Since: 1.6
+ **/
+void
+pango_matrix_translate (PangoMatrix *matrix,
+			double       tx,
+			double       ty)
+{
+  g_return_if_fail (matrix != NULL);
+
+  matrix->x0  = matrix->xx * tx + matrix->xy * ty + matrix->x0;
+  matrix->y0  = matrix->yx * tx + matrix->yy * ty + matrix->y0;
+}
+
+/**
+ * pango_matrix_scale:
+ * @matrix: a #PangoMatrix
+ * @scale_x: amount to scale by in X direction
+ * @scale_y: amount to scale by in Y direction
+ * 
+ * Changes the transformation represented by @matrix to be the
+ * transformation given by first scaling by @sx in the X direction
+ * and @sy in the Y direction then applying the original
+ * transformation.
+ *
+ * Since: 1.6
+ **/
+void
+pango_matrix_scale (PangoMatrix *matrix,
+		    double       scale_x,
+		    double       scale_y)
+{
+  g_return_if_fail (matrix != NULL);
+
+  matrix->xx *= scale_x;
+  matrix->xy *= scale_y;
+  matrix->yx *= scale_x;
+  matrix->yy *= scale_y;
+}
+
+/**
+ * pango_matrix_rotate:
+ * @matrix: a #PangoMatrix
+ * @degrees: degrees to rotate counter-clockwise
+ * 
+ * Changes the transformation represented by @matrix to be the
+ * transformation given by first rotating by @degrees degrees
+ * counter-clokwise then applying the original transformation.
+ *
+ * Since: 1.6
+ **/
+void
+pango_matrix_rotate (PangoMatrix *matrix,
+		     double       degrees)
+{
+  PangoMatrix tmp;
+  gdouble r, s, c;
+
+  g_return_if_fail (matrix != NULL);
+
+  r = degrees * (G_PI / 180.);
+  s = sin (r);
+  c = cos (r);
+
+  tmp.xx = c;
+  tmp.xy = s;
+  tmp.yx = -s;
+  tmp.yy = c;
+  tmp.x0 = 0;
+  tmp.y0 = 0;
+
+  pango_matrix_concat (matrix, &tmp);
+}
+
+/**
+ * pango_matrix_concat:
+ * @matrix: a #PangoMatrix
+ * @new_matrix: a #PangoMatrix
+ * 
+ * Changes the transformation represented by @matrix to be the
+ * transformation given by first applying transformation
+ * given by @new_matrix then applying the original transformation.
+ *
+ * Since: 1.6
+ **/
+void
+pango_matrix_concat (PangoMatrix       *matrix,
+		     const PangoMatrix *new_matrix)
+{
+  PangoMatrix tmp;
+  
+  g_return_if_fail (matrix != NULL);
+
+  tmp = *matrix;
+
+  matrix->xx = tmp.xx * new_matrix->xx + tmp.xy * new_matrix->yx;
+  matrix->xy = tmp.xx * new_matrix->xy + tmp.xy * new_matrix->yy;
+  matrix->yx = tmp.yx * new_matrix->xx + tmp.yy * new_matrix->yx;
+  matrix->yy = tmp.yx * new_matrix->xy + tmp.yy * new_matrix->yy;
+  matrix->x0  = tmp.xx * new_matrix->x0 + tmp.xy * new_matrix->y0 + tmp.x0;
+  matrix->y0  = tmp.yx * new_matrix->y0 + tmp.yy * new_matrix->y0 + tmp.y0;
+}
+
 static const char canon_map[256] = {
    0,   0,   0,   0,   0,   0,   0,   0,    0,   0,   0,   0,   0,   0,   0,   0, 
    0,   0,   0,   0,   0,   0,   0,   0,    0,   0,   0,   0,   0,   0,   0,   0, 
@@ -1165,7 +1326,7 @@ lang_info_compare (const void *key, const void *val)
  * in the following have sufficient text to excercise all the accents for the language, and
  * there are obviously many more languages to include as well.
  */
-static LangInfo lang_texts[] = {
+static const LangInfo lang_texts[] = {
   { "ar", "Arabic  \330\247\331\204\330\263\331\204\330\247\331\205 \330\271\331\204\331\212\331\203\331\205" },
   { "cs", "Czech (\304\215esky)  Dobr\303\275 den" },
   { "da", "Danish (Dansk)  Hej, Goddag" },
@@ -1255,6 +1416,21 @@ pango_log2vis_get_embedding_levels (gunichar       *str,
 
 #endif /* HAVE_FRIBIDI */
 
+/**
+ * pango_get_mirror_char:
+ * @ch: a unicode character
+ * @mirrored_ch: location to store the mirrored character
+ *
+ * If @ch has the Unicode mirrored property and there is another unicode
+ * character that typically has a glyph that is the mirror image of @ch's
+ * glyph, puts that character in the address pointed to by @mirrored_ch.
+ *
+ * Use g_unichar_get_mirror_char() instead; the docs for that function
+ * provide full details.
+ * 
+ * Return value: %TRUE if @ch has a mirrored character and @mirrored_ch is
+ * filled in, %FALSE otherwise
+ **/
 gboolean 
 pango_get_mirror_char (gunichar        ch,
 		       gunichar       *mirrored_ch)
