@@ -31,6 +31,7 @@ typedef struct _PangoFcCoverageKey  PangoFcCoverageKey;
 typedef struct _PangoFcFace         PangoFcFace;
 typedef struct _PangoFcFamily       PangoFcFamily;
 typedef struct _PangoFcPatternSet   PangoFcPatternSet;
+typedef struct _PangoFcFindFuncInfo PangoFcFindFuncInfo;
 
 #define PANGO_FC_TYPE_FAMILY              (pango_fc_family_get_type ())
 #define PANGO_FC_FAMILY(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_FC_TYPE_FAMILY, PangoFcFamily))
@@ -60,6 +61,9 @@ struct _PangoFcFontMapPrivate
   PangoFcFamily **families;
   int n_families;		/* -1 == uninitialized */
 
+  /* Decoders */
+  GSList *findfuncs;
+
   guint closed : 1;
 };
 
@@ -98,11 +102,17 @@ struct _PangoFcPatternSet
   GList *cache_link;
 };
 
+struct _PangoFcFindFuncInfo
+{
+  PangoFcDecoderFindFunc findfunc;
+  gpointer               user_data;
+  GDestroyNotify         dnotify;
+  gpointer               ddata;
+};
+
 static GType    pango_fc_family_get_type     (void);
 static GType    pango_fc_face_get_type       (void);
 
-static void          pango_fc_font_map_init          (PangoFcFontMap              *fontmap);
-static void          pango_fc_font_map_class_init    (PangoFontMapClass            *class);
 static void          pango_fc_font_map_finalize      (GObject                      *object);
 static PangoFont *   pango_fc_font_map_load_font     (PangoFontMap                 *fontmap,
 						       PangoContext                 *context,
@@ -125,35 +135,7 @@ static guint    pango_fc_coverage_key_hash  (PangoFcCoverageKey *key);
 static gboolean pango_fc_coverage_key_equal (PangoFcCoverageKey *key1,
 					      PangoFcCoverageKey *key2);
 
-static PangoFontClass *parent_class;
-
-GType
-pango_fc_font_map_get_type (void)
-{
-  static GType object_type = 0;
-
-  if (!object_type)
-    {
-      static const GTypeInfo object_info =
-      {
-        sizeof (PangoFcFontMapClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) pango_fc_font_map_class_init,
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (PangoFcFontMap),
-        0,              /* n_preallocs */
-        (GInstanceInitFunc) pango_fc_font_map_init,
-      };
-      
-      object_type = g_type_register_static (PANGO_TYPE_FONT_MAP,
-                                            "PangoFcFontMap",
-                                            &object_info, 0);
-    }
-  
-  return object_type;
-}
+G_DEFINE_TYPE (PangoFcFontMap, pango_fc_font_map, PANGO_TYPE_FONT_MAP)
 
 static void 
 pango_fc_font_map_init (PangoFcFontMap *fcfontmap)
@@ -186,17 +168,16 @@ pango_fc_font_map_init (PangoFcFontMap *fcfontmap)
 }
 
 static void
-pango_fc_font_map_class_init (PangoFontMapClass *class)
+pango_fc_font_map_class_init (PangoFcFontMapClass *class)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (class);
-  
-  parent_class = g_type_class_peek_parent (class);
+  PangoFontMapClass *fontmap_class = PANGO_FONT_MAP_CLASS (class);
   
   object_class->finalize = pango_fc_font_map_finalize;
-  class->load_font = pango_fc_font_map_load_font;
-  class->load_fontset = pango_fc_font_map_load_fontset;
-  class->list_families = pango_fc_font_map_list_families;
-  class->shape_engine_type = PANGO_RENDER_TYPE_FC;
+  fontmap_class->load_font = pango_fc_font_map_load_font;
+  fontmap_class->load_fontset = pango_fc_font_map_load_fontset;
+  fontmap_class->list_families = pango_fc_font_map_list_families;
+  fontmap_class->shape_engine_type = PANGO_RENDER_TYPE_FC;
 
   g_type_class_add_private (object_class, sizeof (PangoFcFontMapPrivate));
 }
@@ -330,6 +311,41 @@ pango_fc_clear_pattern_hashes (PangoFcFontMap *fcfontmap)
   priv->fontset_hash_list = NULL;
 }
 
+/**
+ * pango_fc_font_map_add_decoder_find_func:
+ * @fcfontmap: The #PangoFcFontMap to add this method to.
+ * @findfunc: The #PangoFcDecoderFindFunc callback function
+ * @user_data: User data.
+ * @dnotify: A #GDestroyNotify callback that will be called when the
+ *  fontmap is finalized and the decoder is released.
+ *
+ * This function saves a callback method in the #PangoFcFontMap that
+ * will be called whenever new fonts are created.  If the
+ * function returns a #PangoFcDecoder, that decoder will be used to
+ * determine both coverage via a #FcCharSet and a one-to-one mapping of
+ * characters to glyphs.  This will allow applications to have
+ * application-specific encodings for various fonts.
+ *
+ * Since: 1.6.
+ **/
+void
+pango_fc_font_map_add_decoder_find_func (PangoFcFontMap        *fcfontmap,
+					 PangoFcDecoderFindFunc findfunc,
+					 gpointer               user_data,
+					 GDestroyNotify         dnotify)
+{
+  PangoFcFontMapPrivate *priv = fcfontmap->priv;
+  PangoFcFindFuncInfo *info;
+
+  info = g_new (PangoFcFindFuncInfo, 1);
+
+  info->findfunc = findfunc;
+  info->user_data = user_data;
+  info->dnotify = dnotify;
+
+  priv->findfuncs = g_slist_append (priv->findfuncs, info);
+}
+
 static void
 pango_fc_font_map_finalize (GObject *object)
 {
@@ -346,9 +362,19 @@ pango_fc_font_map_finalize (GObject *object)
   if (priv->pattern_hash)
     g_hash_table_destroy (priv->pattern_hash);
 
-  G_OBJECT_CLASS (parent_class)->finalize (object);
-}
+  while (priv->findfuncs)
+    {
+      PangoFcFindFuncInfo *info;
+      info = priv->findfuncs->data;
+      if (info->dnotify)
+	info->dnotify (info->user_data);
 
+      g_free (info);
+      priv->findfuncs = g_slist_delete_link (priv->findfuncs, priv->findfuncs);
+    }
+
+  G_OBJECT_CLASS (pango_fc_font_map_parent_class)->finalize (object);
+}
 
 /* Add a mapping from xfont->font_pattern to xfont */
 static void
@@ -359,7 +385,6 @@ pango_fc_font_map_add (PangoFcFontMap *fcfontmap,
 
   g_assert (fcfont->fontmap == NULL);
 	
-  fcfont->fontmap = g_object_ref (fcfontmap);
   g_hash_table_insert (priv->fonts,
 		       fcfont->font_pattern,
 		       fcfont);
@@ -587,12 +612,15 @@ pango_fc_make_pattern (const PangoFontDescription *description)
 }
 
 static PangoFont *
-pango_fc_font_map_new_font (PangoFontMap  *fontmap,
-			    FcPattern      *match)
+pango_fc_font_map_new_font (PangoFontMap      *fontmap,
+			    const PangoMatrix *pango_matrix,
+			    FcPattern     *match)
 {
   PangoFcFontMap *fcfontmap = (PangoFcFontMap *)fontmap;
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
+  FcPattern *pattern;
   PangoFcFont *fcfont;
+  GSList *l;
 
   /* Returning NULL here actually violates a contract
    * that loading load_font() will never return NULL.
@@ -604,14 +632,57 @@ pango_fc_font_map_new_font (PangoFontMap  *fontmap,
     return NULL;
   
   /* Look up cache */
-  fcfont = g_hash_table_lookup (priv->fonts, match);
+  if (!pango_matrix)
+    {
+      fcfont = g_hash_table_lookup (priv->fonts, match);
   
-  if (fcfont)
-    return g_object_ref (fcfont);
+      if (fcfont)
+	return g_object_ref (fcfont);
+    }
 
-  fcfont = PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->new_font (fcfontmap, match);
+  if (pango_matrix)
+    {
+      FcMatrix fc_matrix;
 
-  pango_fc_font_map_add (fcfontmap, fcfont);
+      /* FontConfig has the Y axis pointing up, Pango, down.
+       */
+      fc_matrix.xx = pango_matrix->xx;
+      fc_matrix.xy = - pango_matrix->xy;
+      fc_matrix.yx = - pango_matrix->yx;
+      fc_matrix.yy = pango_matrix->yy;
+      
+      pattern = FcPatternDuplicate (match);
+      FcPatternAddMatrix (pattern, FC_MATRIX, &fc_matrix);
+    }
+  else
+    pattern = match;
+
+  fcfont = PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->new_font (fcfontmap, pattern);
+
+  if (!pango_matrix)
+    pango_fc_font_map_add (fcfontmap, fcfont);
+
+  if (pango_matrix)
+    FcPatternDestroy (pattern);
+
+  fcfont->fontmap = g_object_ref (fcfontmap);
+
+  /*
+   * Give any custom decoders a crack at this font now that it's been
+   * created.
+   */
+  for (l = priv->findfuncs; l && l->data; l = l->next)
+    {
+      PangoFcFindFuncInfo *info = l->data;
+      PangoFcDecoder *decoder;
+
+      decoder = info->findfunc (match, info->user_data);
+      if (decoder)
+	{
+	  _pango_fc_font_set_decoder (fcfont, decoder);
+	  break;
+	}
+    }
 
   return (PangoFont *)fcfont;
 }
@@ -736,15 +807,19 @@ pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
 
 static PangoFont *
 pango_fc_font_map_load_font (PangoFontMap               *fontmap,
-			      PangoContext               *context,
-			      const PangoFontDescription *description)
+			     PangoContext               *context,
+			     const PangoFontDescription *description)
 {
   PangoFcPatternSet *patterns = pango_fc_font_map_get_patterns (fontmap, context, description, NULL);
   if (!patterns)
     return NULL;
 
   if (patterns->n_patterns > 0)
-    return pango_fc_font_map_new_font (fontmap, patterns->patterns[0]);
+    {
+      return pango_fc_font_map_new_font (fontmap,
+					 context ? pango_context_get_matrix (context) : NULL,
+					 patterns->patterns[0]);
+    }
   
   return NULL;
 }
@@ -809,35 +884,51 @@ pango_fc_font_map_load_fontset (PangoFontMap                 *fontmap,
   PangoFcPatternSet *patterns = pango_fc_font_map_get_patterns (fontmap, context, desc, language);
   PangoFcFontMap *fcfontmap = PANGO_FC_FONT_MAP (fontmap);
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
+  PangoFontset *result;
+  const PangoMatrix *matrix;
   int i;
   
   if (!patterns)
     return NULL;
 
-  if (!patterns->fontset)
+  if (context)
+    matrix = pango_context_get_matrix (context);
+  else
+    matrix = NULL;
+
+  /* We never cache fontsets when a transformation is in place
+   */
+  if (!patterns->fontset || matrix)
     {
       PangoFontsetSimple *simple;
       simple = pango_fontset_simple_new (language);
+      result = PANGO_FONTSET (simple);
       
       for (i = 0; i < patterns->n_patterns; i++)
 	{
-	  PangoFont *font = pango_fc_font_map_new_font (fontmap, patterns->patterns[i]);
+	  PangoFont *font;
+
+	  font = pango_fc_font_map_new_font (fontmap, matrix, patterns->patterns[i]);
 	  if (font)
 	    pango_fontset_simple_append (simple, font);
 	}
-  
-      patterns->fontset = PANGO_FONTSET (simple);
-      g_object_add_weak_pointer (G_OBJECT (patterns->fontset),
-				 (gpointer *)&patterns->fontset);
+
+      if (!matrix)
+	{
+	  patterns->fontset = PANGO_FONTSET (simple);
+	  g_object_add_weak_pointer (G_OBJECT (patterns->fontset),
+				     (gpointer *)&patterns->fontset);
+	}
     }
   else
-    g_object_ref (patterns->fontset);
+    result = g_object_ref (patterns->fontset);
   
-  if (!patterns->cache_link ||
-      patterns->cache_link != priv->fontset_cache->head)
+  if (!matrix &&
+      (!patterns->cache_link ||
+       patterns->cache_link != priv->fontset_cache->head))
     pango_fc_font_map_cache_fontset (fcfontmap, patterns);
 
-  return patterns->fontset;
+  return result;
 }
 
 static void
@@ -901,25 +992,22 @@ pango_fc_font_map_set_coverage (PangoFcFontMap            *fcfontmap,
 
 PangoCoverage *
 _pango_fc_font_map_get_coverage (PangoFcFontMap *fcfontmap,
-				 FcPattern      *pattern)
+				 PangoFcFont    *fcfont)
 {
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
   PangoFcCoverageKey key;
   PangoCoverage *coverage;
-  FcChar32  map[FC_CHARSET_MAP_SIZE];
-  FcChar32  ucs4, pos;
   FcCharSet *charset;
-  int i;
   
   /*
    * Assume that coverage information is identified by
    * a filename/index pair; there shouldn't be any reason
    * this isn't true, but it's not specified anywhere
    */
-  if (FcPatternGetString (pattern, FC_FILE, 0, (FcChar8 **) &key.filename) != FcResultMatch)
+  if (FcPatternGetString (fcfont->font_pattern, FC_FILE, 0, (FcChar8 **) &key.filename) != FcResultMatch)
     return NULL;
 
-  if (FcPatternGetInteger (pattern, FC_INDEX, 0, &key.id) != FcResultMatch)
+  if (FcPatternGetInteger (fcfont->font_pattern, FC_INDEX, 0, &key.id) != FcResultMatch)
     return NULL;
   
   coverage = g_hash_table_lookup (priv->coverage_hash, &key);
@@ -930,9 +1018,34 @@ _pango_fc_font_map_get_coverage (PangoFcFontMap *fcfontmap,
    * Pull the coverage out of the pattern, this
    * doesn't require loading the font
    */
-  if (FcPatternGetCharSet (pattern, FC_CHARSET, 0, &charset) != FcResultMatch)
+  if (FcPatternGetCharSet (fcfont->font_pattern, FC_CHARSET, 0, &charset) != FcResultMatch)
     return NULL;
-  
+
+  coverage = _pango_fc_font_map_fc_to_coverage (charset);
+
+  pango_fc_font_map_set_coverage (fcfontmap, &key, coverage);
+ 
+  return coverage;
+}
+
+/**
+ * _pango_fc_font_map_fc_to_coverage:
+ * @charset: #FcCharSet to convert to a #PangoCoverage object.
+ *
+ * Convert the given #FcCharSet into a new #PangoCoverage object.  The
+ * caller is responsible for freeing the newly created object.
+ *
+ * Since: 1.6
+ **/
+
+PangoCoverage  *
+_pango_fc_font_map_fc_to_coverage (FcCharSet *charset)
+{
+  PangoCoverage *coverage;
+  FcChar32  ucs4, pos;
+  FcChar32  map[FC_CHARSET_MAP_SIZE];
+  int i;
+
   /*
    * Convert an Fc CharSet into a pango coverage structure.  Sure
    * would be nice to just use the Fc structure in place...
@@ -958,7 +1071,7 @@ _pango_fc_font_map_get_coverage (PangoFcFontMap *fcfontmap,
 	    }
 	}
     }
-  
+
   /* Awful hack so Hangul Tone marks get rendered with the same
    * font and in the same run as other Hangul characters. If a font
    * covers the first composed Hangul glyph, then it is declared to cover
@@ -971,8 +1084,6 @@ _pango_fc_font_map_get_coverage (PangoFcFontMap *fcfontmap,
       pango_coverage_set (coverage, 0x302f, PANGO_COVERAGE_EXACT);
     }
 
-  pango_fc_font_map_set_coverage (fcfontmap, &key, coverage);
- 
   return coverage;
 }
 
